(**************************************************************)
(*
 *  Ensemble, (Version 0.61)
 *  Copyright 1999 Cornell University
 *  All rights reserved.
 *
 *  See ensemble/doc/license.txt for further information.
 *)
(**************************************************************)
(**************************************************************)
(* TR_REKEY.ML : switching to a new key using tree structure *)
(* Authors: Zhen Xiao, Mark Hayden, 6/97 *)
(* Note: If the application makes a rekey request when the group 
 * already blocked or during group merge, such request would be 
 * dropped with no notification. Thus this protocol is a best-effort
 * protocol.
 *)
(**************************************************************)
open Util
open Layer
open View
open Event
open Trans
(**************************************************************)
let name = Trace.filel "TR_REKEY"
(**************************************************************)

type header =
  | Ack
  | Rekey
  | Ticket of rank * Auth.ticket

(**************************************************************)

type state = {
  degree : int ;
  children : rank list ;
  mutable key_sug : Security.key ;
  rekeying : Once.t ;
  blocking : Once.t ;
  ack : bool array
} 

(**************************************************************)

let dump = Layer.layer_dump name (fun (ls,vs) s -> [|
  eprintf "TR_REKEY:%s\n" ls.name ;
  sprintf "ack=%s\n" (string_of_bool_array s.ack) ;
  sprintf "blocking=%s, rekeying=%s\n"
    (Once.to_string s.blocking) (Once.to_string s.rekeying)
)

(**************************************************************)

let init _ (ls,vs) = 
  let degree = Param.int vs.params "tr_rekey_degree" in
  let first_child n = degree * n + 1 in
  let children =
    let rec loop rank degree =
      if (degree = 0) || (rank >= ls.nmembers) then []
      else rank :: (loop (succ rank) (degree - 1))
    in loop (first_child ls.rank) degree
  in
{ 
  degree = degree ;
  children = children ;
  key_sug = vs.key ;
  rekeying = Once.create name ;
  blocking = Once.create "blocking" ;
  ack = array_create name ls.nmembers false
}

(**************************************************************)

let hdlrs s ((ls,vs) as vf) {up_out=up;upnm_out=upnm;dn_out=dn;dnlm_out=dnlm;dnnm_out=dnnm} =
  let failwith = layer_fail dump vf s name in
  let ack = make_acker name dnnm in
  let log = Trace.log name ls.name in

  let all_acked () =
    array_for_all ident s.ack
  in

  (* Encrypt and then send the key to all its children.
   *)
  let dis_key key children =
    log (fun () -> sprintf "Distributing new key to my children: %s" (string_of_int_list s.children)) ;    
    List.iter (fun rank ->
      let dst = Arrayf.get vs.address rank in
      match Auth.ticket ls.addr dst key with
      | Some t -> 
	  (* Use tree structure to distributing the key.
	   *)
	  dnlm (sendPeer name rank) (Ticket(ls.rank,t))
      | None -> 
	  (* We suspect this member.
	   *)
	  log (fun () -> sprintf "I didn't authenticate member %d" rank) ;
	  dnnm (suspectReason name (Arrayf.of_ranks ls.nmembers [rank]) "TR_REKEY:can create ticket")
    ) children
  in

  let try_rekey () =
    if not (Once.isset s.rekeying) 
    && not (Once.isset s.blocking)
    then (
      log (fun () -> sprintf "Coordinator rekeying: begin") ;
      Once.set s.rekeying ;

      (* Generate a new shared key.
       *)
      s.key_sug <- Security.create () ;
      let key = Security.str_of_key s.key_sug in
      log (fun () -> sprintf "generating new keys:%s" (Security.string_of_key s.key_sug)) ;
      dis_key key s.children ;
      s.ack.(ls.rank) <- true ;
      dnlm (castEv name) Ack
      ; log (fun () -> sprintf "Coordinator rekeying: end")
    )
  in


  let forward_key key =
    if not (Once.isset s.rekeying) 
    && not (Once.isset s.blocking)
    then (
      Once.set s.rekeying ;
      let key = Security.str_of_key s.key_sug in
      dis_key key s.children ;
    )
  in


  let up_hdlr ev abv () = up ev abv

  and uplm_hdlr ev hdr = match getType ev,hdr with

    (* Got a request to start rekeying the group.
     *)
  | (ECast|ESend), Rekey ->
      if vs.coord = ls.rank then
      	try_rekey () ;
      (*ack ev ;*) free name ev
      
    (* Got a ticket from the parent.  Try to decrypt it. If succeed, 
     * broadcast [Ack] to the group and remember the key.  
     *)
  | (ECast|ESend), Ticket(origin,t) ->
      if (not ls.am_coord)
      && not (Once.isset s.blocking) 
      then (
	match Auth.check ls.addr t with
	| Some(princ,key) ->
	    if Arrayf.mem princ (Addr.array_of_set (Arrayf.get vs.address origin)) then (
	      s.key_sug <- Security.Common(key) ;
	      log (fun () -> sprintf "receiving new keys:%s" (Security.string_of_key s.key_sug)) ;
	      log (fun () -> sprintf "got the new key successfully from member %d" origin) ;
	      s.ack.(origin) <- true ;
	      s.ack.(ls.rank) <- true ;
	      forward_key key ;
	      dnlm (castEv name) Ack
	    ) else (
	      (* We suspect them if the origin is invalid.
	       *)
	      log (fun () -> sprintf "got the new key successfully, but invalid addr") ;
	      dnnm (suspectReason name (Arrayf.of_ranks ls.nmembers [origin]) "TR_REKEY:can't auth ticket")
	    )
	| None -> 
	    (* We suspect them if we can't auth ticket.
	     *)
	    log (fun () -> sprintf "member didn't authenticate its parent") ;
	    dnnm (suspectReason name (Arrayf.of_ranks ls.nmembers [origin]) "TR_REKEY:can't auth ticket")
      ) ;
      (*ack ev ;*) free name ev 

  | (ECast|ESend), Ack ->
      (* Some members replied that they got the key. Check whether
       * all members have the key. If so, prompt for a view change.
       *)
      let origin = getPeer ev in 
      s.ack.(origin) <- true ;

      if vs.coord = ls.rank && all_acked () then (
        log (fun () -> sprintf "%s members got the key" (string_of_int ls.nmembers)) ;
	  dnnm (create name EPrompt[]) ;
      );
      (*ack ev ;*) free name ev 

  | _ -> failwith unknown_local

  and upnm_hdlr ev = match getType ev with

    (* Switch to a new conversation key.  Here we assume the
     * Auth module authenticated a member when it did
     * encryption for that member.  
     *)
  | ERekey ->
      log (fun () -> sprintf "ERekey event") ;
      if ls.rank = vs.coord then (
	try_rekey ()
      ) else (
      	dnlm (sendPeer name vs.coord) Rekey
      ) ;
      upnm ev

    (* EBlock: Mark that we are now blocking.
     *)
  | EBlock ->
      Once.set s.blocking ;
      upnm ev

    (* EFail: Mark failed members as acked, since there is
     * no need to wait for failed members.  
     *)
  | EFail ->      
      let failed = getFailures ev in
      for rank = 0 to pred ls.nmembers do
	if Arrayf.get failed rank then (
	  log (fun () -> sprintf "member %d failed" rank);
	  s.ack.(rank) <- true
	)
      done ;

      (* Prompt a new view change if all remaining members are
       * now ready.
       *)
      if vs.coord = ls.rank && all_acked () then 
	dnnm (create name EPrompt[]) ;
      upnm ev

    (* EView: If all members have acknowledged the new key and no group 
     * merge occurs, then switch to using it.
     *)
  | EView ->
      let view_old = Arrayf.to_list vs.view in
      let vs_new = getViewState ev in
      let view_new = Arrayf.to_list vs_new.view in
      let no_merger = Lset.subtract view_new view_old = [] in
      if no_merger && all_acked () then (
	log (fun () -> sprintf "EView:switching keys:%s" (Security.string_of_key s.key_sug)) ;
	let vs = getViewState ev in
	let vs = View.set vs [Vs_key s.key_sug] in
	let ev = set name ev[ViewState vs] in
      	upnm ev
      ) else (
	log (fun () -> sprintf "My ack array:%s" (string_of_bool_array s.ack)) ;
	upnm ev
      )

  | EDump -> ( dump vf s ; upnm ev )
  | _ -> upnm ev

  and dn_hdlr ev abv = dn ev abv ()

  and dnnm_hdlr = dnnm

in {up_in=up_hdlr;uplm_in=uplm_hdlr;upnm_in=upnm_hdlr;dn_in=dn_hdlr;dnnm_in=dnnm_hdlr}

let l args vf = Layer.hdr init hdlrs None NoOpt args vf

let _ = 
  Param.default "tr_rekey_degree" (Param.Int 2) ;
  Elink.layer_install name l

(**************************************************************)
